"""
Tests of macro_params.py module
"""

from ogphl import macro_params


def test_get_macro_params():
    test_dict = macro_params.get_macro_params()

    assert isinstance(test_dict, dict)
    assert (
        list(test_dict.keys()).sort()
        == [
            "r_gov_shift",
            "r_gov_scale",
            "alpha_T",
            "alpha_G",
            "initial_debt_ratio",
            "g_y_annual",
            "gamma",
            "zeta_D",
            "initial_foreign_debt_ratio",
        ].sort()
    )



"""
Tests of input_output.py module
"""

import pandas as pd
import pytest
from ogphl import input_output as io


sam_dict = {
    # "index": ["Beer", "Chocolate", "Car", "House"],
    "Ag": [30, 160, 0, 5],
    "Mining": [10, 0, 100, 100],
    "Manufacturing": [60, 40, 200, 295],
    "hhd-r1": [100, 200, 300, 400],
    "hhd-r2": [0, 0, 0, 0],
    "hhd-r3": [0, 0, 0, 0],
    "hhd-r4": [0, 0, 0, 0],
    "hhd-r5": [0, 0, 0, 0],
    "hhd-u1": [0, 0, 0, 0],
    "hhd-u2": [0, 0, 0, 0],
    "hhd-u3": [0, 0, 0, 0],
    "hhd-u4": [0, 0, 0, 0],
    "hhd-u5": [0, 0, 0, 0],
    "row": [10, 20, 30, 40],
}
sam_df = pd.DataFrame(sam_dict, index=["Beer", "Chocolate", "Car", "House"])

cons_dict = {"Food": ["Beer", "Chocolate"], "Non-food": ["Car", "House"]}

prod_dict = {
    "Primary": ["Ag", "Mining"],
    "Secondary": ["Manufacturing"],
}


@pytest.mark.parametrize(
    "sam_df, cons_dict",
    [
        (sam_df, cons_dict),
    ],
    ids=["Test 1"],
)
def test_get_alpha_c(sam_df, cons_dict):
    """
    Test of get_alpha_c() function
    """
    test_dict = io.get_alpha_c(sam=sam_df, cons_dict=cons_dict)

    assert isinstance(test_dict, dict)
    assert list(test_dict.keys()).sort() == ["Food", "Non-food"].sort()
    assert test_dict["Food"] == 270 / 900
    assert test_dict["Non-food"] == 630 / 900


@pytest.mark.parametrize(
    "sam_df, cons_dict, prod_dict",
    [
        (sam_df, cons_dict, prod_dict),
    ],
    ids=["Test 1"],
)
def test_get_io_matrix(sam_df, cons_dict, prod_dict):
    """
    Test of get_io_matrix()
    """
    test_df = io.get_io_matrix(
        sam=sam_df, cons_dict=cons_dict, prod_dict=prod_dict
    )

    assert isinstance(test_df, pd.DataFrame)
    assert list(test_df.columns).sort() == ["Primary", "Secondary"].sort()
    assert list(test_df.index).sort() == ["Food", "Non-food"].sort()
    assert test_df.loc["Food", "Primary"] == 2 / 3
    assert test_df.loc["Food", "Secondary"] == 1 / 3



"""
This model tests whether using the `OG-PHL/examples/run_og_phl.py`
work by making sure that it does not break (is still running) after
5 minutes (300 seconds).
"""

import multiprocessing
import time
import os
import sys
import importlib.util
from pathlib import Path
import pytest


def call_run_og_phl():
    cur_path = os.path.split(os.path.abspath(__file__))[0]
    path = Path(cur_path)
    roe_fldr = os.path.join(path, "..", "examples")
    roe_file_path = os.path.join(roe_fldr, "run_og_phl.py")
    spec = importlib.util.spec_from_file_location(
        "run_og_phl.py", roe_file_path
    )
    roe_module = importlib.util.module_from_spec(spec)
    sys.modules["run_og_phl.py"] = roe_module
    spec.loader.exec_module(roe_module)
    roe_module.main()


@pytest.mark.local
def test_run_og_phl(f=call_run_og_phl):
    p = multiprocessing.Process(target=f, name="run_og_phl", args=())
    p.start()
    time.sleep(300)
    if p.is_alive():
        p.terminate()
        p.join()
        timetest = True
    else:
        print("run_og_phl.py did not run for minimum time")
        timetest = False
    print("timetest ==", timetest)

    assert timetest



"""
This module uses data from World Bank WDI, World Bank Quarterly Public
Sector Debt (QPSD) database, the IMF, and UN ILO to find values for
parameters for the OG-PHL model that rely on macro data for calibration.
"""

# imports
from pandas_datareader import wb
import pandas as pd
import numpy as np
import requests
import datetime
import statsmodels.api as sm
from io import StringIO


def get_macro_params(
    data_start_date=datetime.datetime(1947, 1, 1),
    data_end_date=datetime.datetime(2023, 1, 1),
    country_iso="PHL",
):
    """
    Compute values of parameters that are derived from macro data

    Args:
        data_start_date (datetime): start date for data
        data_end_date (datetime): end date for data
        country_iso (str): ISO code for country

    Returns:
        macro_parameters (dict): dictionary of parameter values
    """
    # initialize a dictionary of parameters
    macro_parameters = {}

    """
    Retrieve data from the World Bank World Development Indicators.
    """
    # Dictionaries of variables and their corresponding World Bank codes
    # Annual data
    wb_a_variable_dict = {
        "GDP per capita (constant 2015 US$)": "NY.GDP.PCAP.KD",
        "Real GDP (constant 2015 US$)": "NY.GDP.MKTP.KD",
        "Nominal GDP (current US$)": "NY.GDP.MKTP.CD",
        "General government final consumption expenditure (current US$)": "NE.CON.GOVT.CD",
        "External debt stocks, public and publicly guaranteed (PPG) (DOD, current US$)": "DT.DOD.DPPG.CD",
        "External debt stocks, total (DOD, current US$)": "DT.DOD.DECT.CD",
        r"External debt stocks (% of GNI)": "DT.DOD.DECT.GN.ZS",
        r"Central government debt, total (% of GDP)": "GC.DOD.TOTL.GD.ZS",
        r"General government final consumption expenditure (% of GDP)": "NE.CON.GOVT.ZS",
    }
    try:
        # pull series of interest from the WB using pandas_datareader
        # Annual data
        wb_data_a = wb.WorldBankReader(
            symbols=wb_a_variable_dict.values(),
            countries=country_iso,
            start=data_start_date,
            end=data_end_date,
            freq="A",
        ).read()
        wb_data_a.rename(
            columns=dict((y, x) for x, y in wb_a_variable_dict.items()),
            inplace=True,
        )
        # Remove the hierarchical index (country and year) of
        # wb_data_a and create a single row index using year
        wb_data_a.reset_index(inplace=True)
        wb_data_a["year"] = wb_data_a.year.astype(int)
        wb_data_a = wb_data_a.set_index("year")
        # Compute macro parameters from WB data
        # Latest from WB WDI data is 2014, so use: https://www.treasury.gov.ph/?p=64737
        macro_parameters["initial_debt_ratio"] = 0.60
        macro_parameters["initial_foreign_debt_ratio"] = (
            pd.Series(
                wb_data_a[r"External debt stocks (% of GNI)"]
                * (
                    wb_data_a[
                        "External debt stocks, public and publicly guaranteed (PPG) (DOD, current US$)"
                    ]
                    / wb_data_a[
                        "External debt stocks, total (DOD, current US$)"
                    ]
                )
            ).loc[data_end_date.year - 2]
            / 100
        )
        # zeta_D = share of new debt issues from government that are
        # purchased by foreigners
        # set to initial ratio without better info
        macro_parameters["zeta_D"] = [
            macro_parameters["initial_foreign_debt_ratio"]
        ]
        macro_parameters["g_y_annual"] = (
            wb_data_a["GDP per capita (constant 2015 US$)"]
            .loc[2000:2019]  # stop pre-COVID
            .pct_change()
            .mean()
        )
        macro_parameters["alpha_G"] = [
            (
                wb_data_a[
                    r"General government final consumption expenditure (% of GDP)"
                ].loc[data_end_date.year]
                / 100
            )
        ]
    except:
        print("Failed to retrieve data from World Bank")
        print("Will not update the following parameters:")
        print(
            "[initial_debt_ratio, initial_foreign_debt_ratio, zeta_D, g_y, alpha_G]"
        )

    """
    Retrieve labour share data from the United Nations ILOSTAT Data API
    (see https://rshiny.ilo.org/dataexplorer9/?lang=en)
    """
    target = (
        "https://rplumber.ilo.org/data/indicator/"
        + "?id=LAP_2GDP_NOC_RT_A"
        + "&ref_area="
        + str(country_iso)
        + "&timefrom="
        + str(data_start_date.year)
        + "&timeto="
        + str(data_end_date.year)
        + "&type=both&format=.csv"
    )
    print("ILO data target = ", target)
    response = requests.get(target, timeout=30)
    if response.status_code == 200:
        csv_content = StringIO(response.text)
        df_temp = pd.read_csv(csv_content)
    else:
        print(
            f"Failed to retrieve data. HTTP status code: {response.status_code}"
        )
    ilo_data = df_temp[["time", "obs_value"]]
    # find gamma, capital's share of income
    macro_parameters["gamma"] = [
        1
        - (
            (
                ilo_data.loc[
                    ilo_data["time"] == min(data_end_date.year, 2021),
                    "obs_value",  # 2021 is latest year of data
                ].squeeze()
            )
            / 100
        )
    ]

    """
    Calibrate parameters from IMF data
    """
    # alpha_T, non-social security benefits as a fraction of GDP
    # can't find this specifically, so use primary expenditures minus
    # final consumption expenditures
    # source: https://www.imf.org/external/datamapper/profile/PHL
    macro_parameters["alpha_T"] = [0.2391 - 0.142]

    """"
    Esimate the discount on sovereign yields relative to private debt
    Follow the methodology in Li, Magud, Werner, Witte (2021)
    available at:
    https://www.imf.org/en/Publications/WP/Issues/2021/06/04/The-Long-Run-Impact-of-Sovereign-Yields-on-Corporate-Yields-in-Emerging-Markets-50224

    Steps:
    1) Generate modelled corporate yields (corp_yhat) for a range of
    sovereign yields (sov_y)  using the estimated equation in col 2 of
    table 8 (and figure 3). 2) Estimate the OLS using sovereign yields
    as the dependent variable
    """

    # # estimate r_gov_shift and r_gov_scale
    sov_y = np.arange(20, 120) / 10
    corp_yhat = 8.199 - (2.975 * sov_y) + (0.478 * sov_y**2)
    corp_yhat = sm.add_constant(corp_yhat)
    mod = sm.OLS(
        sov_y,
        corp_yhat,
    )
    res = mod.fit()
    # First term is the constant and needs to be divided by 100 to have
    # the correct unit. Second term is the coefficient
    macro_parameters["r_gov_shift"] = [-res.params[0] / 100]
    macro_parameters["r_gov_scale"] = [res.params[1]]

    return macro_parameters



import pandas as pd
import numpy as np
import os
from ogphl.constants import CONS_DICT, PROD_DICT


CUR_DIR = os.path.dirname(os.path.realpath(__file__))
"""
Read in Social Accounting Matrix (SAM) file
"""
# Read in SAM file
# SAM file:
sam_path = os.path.join(CUR_DIR, "data", "002_IFPRI_SAM_PHL_2018_SAM.csv")
SAM = pd.read_csv(sam_path, index_col=1, thousands=",")
# replace NaN with 0
SAM.fillna(0, inplace=True)


def get_alpha_c(sam=SAM, cons_dict=CONS_DICT):
    """
    Calibrate the alpha_c vector, showing the shares of household
    expenditures for each consumption category

    Args:
        sam (pd.DataFrame): SAM file
        cons_dict (dict): Dictionary of consumption categories

    Returns:
        alpha_c (dict): Dictionary of shares of household expenditures
    """
    hh_cols = [
        "hhd-r1",
        "hhd-r2",
        "hhd-r3",
        "hhd-r4",
        "hhd-r5",
        "hhd-u1",
        "hhd-u2",
        "hhd-u3",
        "hhd-u4",
        "hhd-u5",
    ]
    alpha_c = {}
    overall_sum = 0
    for key, value in cons_dict.items():
        # note the subtraction of the row to focus on domestic consumption
        category_total = (
            sam.loc[sam.index.isin(value), hh_cols].values.astype(float).sum()
        )
        alpha_c[key] = category_total
        overall_sum += category_total
    for key, value in cons_dict.items():
        alpha_c[key] = alpha_c[key] / overall_sum

    return alpha_c


def get_io_matrix(sam=SAM, cons_dict=CONS_DICT, prod_dict=PROD_DICT):
    """
    Calibrate the io_matrix array.  This array relates the share of each
    production category in each consumption category

    Args:
        sam (pd.DataFrame): SAM file
        cons_dict (dict): Dictionary of consumption categories
        prod_dict (dict): Dictionary of production categories

    Returns:
        io_df (pd.DataFrame): Dataframe of io_matrix
    """
    # Create initial matrix as dataframe of 0's to fill in
    io_dict = {}
    for key in prod_dict.keys():
        io_dict[key] = np.zeros(len(cons_dict.keys()))
    io_df = pd.DataFrame(io_dict, index=cons_dict.keys())
    # Fill in the matrix
    # Note, each cell in the SAM represents a payment from the columns
    # account to the row account
    # (see https://www.un.org/en/development/desa/policy/capacity/presentations/manila/6_sam_mams_philippines.pdf)
    # We are thus going to take the consumption categories from rows and
    # the production categories from columns
    for ck, cv in cons_dict.items():
        for pk, pv in prod_dict.items():
            io_df.loc[io_df.index == ck, pk] = (
                sam.loc[sam.index.isin(cv), pv].values.astype(float).sum()
            )
    # change from levels to share (where each row sums to one)
    io_df = io_df.div(io_df.sum(axis=1), axis=0)

    return io_df



from ogphl import macro_params, income
from ogphl import input_output as io
import os
import numpy as np
import datetime
from ogcore import demographics

UN_COUNTRY_CODE = "608"


class Calibration:
    """OG-PHL calibration class"""

    def __init__(
        self,
        p,
        macro_data_start_year=datetime.datetime(1947, 1, 1),
        macro_data_end_year=datetime.datetime(2023, 1, 1),
        demographic_data_path=None,
        output_path=None,
    ):
        """
        Constructor for the Calibration class.

        Args:
            p (OG-Core Specifications object): model parameters
            demographic_data_path (str): path to save demographic data
            output_path (str): path to save output to

        Returns:
            None

        """
        # Create output_path if it doesn't exist
        if output_path is not None:
            if not os.path.exists(output_path):
                os.makedirs(output_path)

        # Macro estimation
        self.macro_params = macro_params.get_macro_params(
            macro_data_start_year, macro_data_end_year
        )

        # io matrix and alpha_c
        if p.I > 1:  # no need if just one consumption good
            alpha_c_dict = io.get_alpha_c()
            # check that model dimensions are consistent with alpha_c
            assert p.I == len(list(alpha_c_dict.keys()))
            self.alpha_c = np.array(list(alpha_c_dict.values()))
        else:
            self.alpha_c = np.array([1.0])
        if p.M > 1:  # no need if just one production good
            io_df = io.get_io_matrix()
            # check that model dimensions are consistent with io_matrix
            assert p.M == len(list(io_df.keys()))
            self.io_matrix = io_df.values
        else:
            self.io_matrix = np.array([[1.0]])

        # demographics
        self.demographic_params = demographics.get_pop_objs(
            p.E,
            p.S,
            p.T,
            0,
            99,
            country_id=UN_COUNTRY_CODE,
            initial_data_year=p.start_year - 1,
            final_data_year=p.start_year + 1,
            GraphDiag=False,
            download_path=demographic_data_path,
        )

        # demographics for 80 period lives (needed for getting e below)
        demog80 = demographics.get_pop_objs(
            20,
            80,
            p.T,
            0,
            99,
            country_id=UN_COUNTRY_CODE,
            initial_data_year=p.start_year - 1,
            final_data_year=p.start_year + 1,
            GraphDiag=False,
        )

        # earnings profiles
        self.e = income.get_e_interp(
            p.E,
            p.S,
            p.J,
            p.lambdas,
            demog80["omega_SS"],
        )

    # method to return all newly calibrated parameters in a dictionary
    def get_dict(self):
        dict = {}
        dict.update(self.macro_params)
        dict["e"] = self.e
        dict["alpha_c"] = self.alpha_c
        dict["io_matrix"] = self.io_matrix
        dict.update(self.demographic_params)

        return dict



SHOW_RUNTIME = False  # Flag to display RuntimeWarnings when run model

REFORM_DIR = "OUTPUT_REFORM"
BASELINE_DIR = "OUTPUT_BASELINE"

# Default year for model runs
DEFAULT_START_YEAR = 2025


VAR_LABELS = {
    "Y": "GDP ($Y_t$)",
    "C": "Consumption ($C_t$)",
    "L": "Labor ($L_t$)",
    "G": "Government Expenditures ($G_t$)",
    "TR": "Lump sum transfers ($TR_t$)",
    "B": "Wealth ($B_t$)",
    "I_total": "Investment ($I_t$)",
    "K": "Capital Stock ($K_t$)",
    "K_d": "Domestically-owned Capital Stock ($K^d_t$)",
    "K_f": "Foreign-owned Capital Stock ($K^f_t$)",
    "D": "Government Debt ($D_t$)",
    "D_d": "Domestically-owned Gov Debt ($D^d_t$)",
    "D_f": "Foreign-owned Gov Debt ($D^f_t$)",
    "r": "Real interest rate ($r_t$)",
    "r_gov": "Real interest rate on gov debt ($r_{gov,t}$)",
    "r_hh": "Real interest rate on HH portfolio ($r_{hh,t}$)",
    "w": "Wage rate",
    "BQ": "Aggregate bequests ($BQ_{j,t}$)",
    "total_tax_revenue": "Total tax revenue ($REV_t$)",
    "business_tax_revenue": "Business tax revenue",
    "iit_revenue": "Individual income tax revenue",
    "payroll_tax_revenue": "Payroll tax revenue",
    "iit_payroll_tax_revenue": "IIT and payroll tax revenue",
    "n_mat": "Labor Supply ($n_{j,s,t}$)",
    "c_path": "Consumption ($c_{j,s,t}$)",
    "bmat_splus1": "Savings ($b_{j,s+1,t+1}$)",
    "bq_path": "Bequests ($bq_{j,s,t}$)",
    "bmat_s": "Savings ($b_{j,s,t}$)",
    "y_before_tax_mat": "Before tax income",
    "etr_path": "Effective Tax Rate ($ETR_{j,s,t}$)",
    "mtrx_path": "Marginal Tax Rate, Labor Income ($MTRx_{j,s,t}$)",
    "mtry_path": "Marginal Tax Rate, Capital Income ($MTRy_{j,s,t}$)",
    "tax_path": "Total Taxes",
    "nssmat": "Labor Supply ($\\bar{n}_{j,s}$)",
    "bssmat_s": "Savings ($\\bar{b}_{j,s}$)",
    "bssmat_splus1": "Savings ($\\bar{b}_{j,s+1}$)",
    "cssmat": "Consumption ($\\bar{c}_{j,s}$)",
    "yss_before_tax_mat": "Before-tax Income",
    "etr_ss": "Effective Tax Rate ($\\bar{ETR}_{j,s}$)",
    "mtrx_ss": "Marginal Tax Rate, Labor Income ($\\bar{MTRx}_{j,s}$)",
    "mtry_ss": "Marginal Tax Rate, Capital Income ($\\bar{MTRy}_{j,s}$)",
    "ETR": "Effective Tax Rates",
    "MTRx": "Marginal Tax Rates on Labor Income",
    "MTRy": "Marginal Tax Rates on Capital Income",
    "etr": "Effective Tax Rates",
    "mtrx": "Marginal Tax Rates on Labor Income",
    "mtry": "Marginal Tax Rates on Capital Income",
    "Yss": "GDP ($\\bar{Y}$)",
    "Css": "Consumption ($\\bar{C}$)",
    "Lss": "Labor ($\\bar{L}$)",
    "Gss": "Government Expenditures ($\\bar{G}$)",
    "TR_ss": "Lump sum transfers, ($\\bar{TR}$)",
    "Bss": "Wealth ($\\bar{B}$)",
    "Iss_total": "Investment ($\\bar{I}$)",
    "Kss": "Capital Stock ($\\bar{K}$)",
    "K_d_ss": "Domestically-owned Capital Stock ($\\bar{K}^d$)",
    "K_f_ss": "Foreign-owned Capital Stock ($\\bar{K}^f$)",
    "Dss": "Government Debt ($\\bar{D}$)",
    "D_d_ss": "Domestically-owned Gov Debt ($\\bar{D}^d$)",
    "D_f_ss": "Foreign-owned Gov Debt ($\\bar{D}^f$)",
    "rss": "Real interest rate ($\\bar{r}$)",
    "r_gov_ss": "Real interest rate on gov debt ($\\bar{r}_{gov}$)",
    "r_hh_ss": "Real interest rate on HH portfolio ($\\bar{r}_{hh}$)",
    "wss": "Wage rate ($\\bar{w}$)",
    "BQss": "Aggregate bequests ($\\bar{BQ}_{j}$)",
    "debt_service_ss": "Debt service cost ($\\bar{r}_{gov}\\bar{D}$)",
    "D/Y": "Debt to GDP ratio",
    "T_Pss": "Government Pensions",
}

ToGDP_LABELS = {
    "D": "Debt-to-GDP ($D_{t}/Y_t$)",
    "D_d": "Domestically-owned Debt-to-GDP ($D^d_{t}/Y_t$)",
    "D_f": "Foreign-owned Debt-to-GDP ($D^f_{t}/Y_t$)",
    "G": "Govt Spending-to-GDP ($G_{t}/Y_t$)",
    "K": "Capital-Output Ratio ($K_{t}/Y_t$)",
    "K_d": "Domestically-owned Capital-Output Ratio ($K^d_{t}/Y_t$)",
    "K_f": "Foreign-owned Capital-Output Ratio ($K^f_{t}/Y_t$)",
    "C": "Consumption-Output Ratio ($C_{t}/Y_t$)",
    "I": "Investment-Output Ratio ($I_{t}/Y_t$)",
    "total_tax_revenue": "Tax Revenue-to-GDP ($REV_{t}/Y_t$)",
}

GROUP_LABELS = {
    7: {
        0: "0-25%",
        1: "25-50%",
        2: "50-70%",
        3: "70-80%",
        4: "80-90%",
        5: "90-99%",
        6: "Top 1%",
    },
    9: {
        0: "0-25%",
        1: "25-50%",
        2: "50-70%",
        3: "70-80%",
        4: "80-90%",
        5: "90-99%",
        6: "99-99.5%",
        7: "99.5-99.9%",
        8: "Top 0.1%",
    },
    10: {
        0: "0-25%",
        1: "25-50%",
        2: "50-70%",
        3: "70-80%",
        4: "80-90%",
        5: "90-99%",
        6: "99-99.5%",
        7: "99.5-99.9%",
        8: "99.9-99.99%",
        9: "Top 0.01%",
    },
}

CBO_UNITS = {
    "Y": r"Billions of ₱",
    "r": "Percent",
    "w_growth": "Percent",
    "L_growth": "Percent",
    "I_total": r"Billions of ₱",
    "L": "2012=100",
    "C": r"Billions of ₱",
    "agg_pension_outlays": r"Billions of ₱",
    "G": r"Billions of ₱",
    "iit_revenue": r"Billions of ₱",
    "payroll_tax_revenue": r"Billions of ₱",
    "business_tax_revenue": r"Billions of ₱",
    "wL": r"Billions of ₱",
    "D": r"Billions of ₱",
}

PARAM_LABELS = {
    "start_year": ["Initial year", r"$\texttt{start_year}$"],
    # 'Gamma': ['Initial distribution of savings', r'\hat{\Gamma}_{0}'],
    # 'N': ['Initial population', 'N_{0}'],
    "omega": ["Population by age over time", r"${{\omega_{s,t}}}_{s=1}^{S}$"],
    # 'fert_rates': ['Fertility rates by age',
    #                r'\left{f_{s}\right}_{s=1}^{S}'],
    "imm_rates": ["Immigration rates by age", r"${{i_{s}}}_{s=1}^{S}$"],
    "rho": ["Mortality rates by age", r"${{\rho_{s}}}_{s=1}^{S}$"],
    "e": ["Deterministic ability process", r"${{e_{j,s}}}_{j,s=1}^{J,S}$"],
    "lambdas": [
        "Lifetime income group percentages",
        r"${{\lambda_{j}}}_{j=1}^{J}$",
    ],
    "J": ["Number of lifetime income groups", "$J$"],
    "S": ["Maximum periods in economically active individual life", "$S$"],
    "E": ["Number of periods of youth economically outside the model", "$E$"],
    "T": ["Number of periods to steady-state", "$T$"],
    "retirement_age": ["Retirement age", "$R$"],
    "ltilde": ["Maximum hours of labor supply", r"$\tilde{l}$"],
    "beta": ["Discount factor", r"$\beta$"],
    "sigma": ["Coefficient of constant relative risk aversion", r"$\sigma$"],
    "frisch": ["Frisch elasticity of labor supply", r"$\nu$"],
    "b_ellipse": ["Scale parameter in utility of leisure", "$b$"],
    "upsilon": ["Shape parameter in utility of leisure", r"$\upsilon$"],
    # 'k': ['Constant parameter in utility of leisure', 'k'],
    "chi_n": [
        "Disutility of labor level parameters",
        r"$\left{\chi^{n}_{s}\right}_{s=1}^{S}$",
    ],
    "chi_b": [
        "Utility of bequests level parameters",
        r"$\left{\chi^{b}_{j}\right}_{j=1}^{J}$",
    ],
    "use_zeta": [
        "Whether to distribute bequests between lifetime income groups",
        r"$\texttt{use_zeta}$",
    ],
    "zeta": ["Distribution of bequests", r"$\zeta$"],
    "Z": ["Total factor productivity", "$Z_{t}$"],
    "gamma": ["Capital share of income", r"$\gamma$"],
    "epsilon": [
        "Elasticity of substitution between capital and labor",
        r"\varepsilon",
    ],
    "delta": ["Capital depreciation rate", r"$\delta$"],
    "g_y": [
        "Growth rate of labor augmenting technological progress",
        r"$g_{y}$",
    ],
    "tax_func_type": [
        "Functional form used for income tax functions",
        r"$\texttt{tax_func_type}$",
    ],
    "analytical_mtrs": [
        "Whether use analytical MTRs or estimate MTRs",
        r"$\texttt{analytical_mtrs}$",
    ],
    "age_specific": [
        "Whether use age-specific tax functions",
        r"$\texttt{age_specific}$",
    ],
    "tau_payroll": ["Payroll tax rate", r"$\tau^{p}_{t}$"],
    # 'theta': ['Replacement rate by average income',
    #           r'\left{\theta_{j}\right}_{j=1}^{J}'],
    "tau_bq": ["Bequest (estate) tax rate", r"$\tau^{BQ}_{t}}$"],
    "tau_b": ["Entity-level business income tax rate", r"$\tau^{b}_{t}$"],
    "delta_tau": [
        "Rate of depreciation for tax purposes",
        r"$\delta^{\tau}_{t}$",
    ],
    "tau_c": ["Consumption tax rates", r"$\tau^{c}_{t,s,j}$"],
    "h_wealth": ["Coefficient on linear term in wealth tax function", "$H$"],
    "m_wealth": ["Constant in wealth tax function", "$M$"],
    "p_wealth": ["Coefficient on level term in wealth tax function", "$P$"],
    "budget_balance": [
        "Whether have a balanced budget in each period",
        r"$\texttt{budget_balance}$",
    ],
    "baseline_spending": [
        "Whether level of spending constant between "
        + "the baseline and reform runs",
        r"$\texttt{baseline_spending}$",
    ],
    "alpha_T": ["Transfers as a share of GDP", r"$\alpha^{T}_{t}$"],
    "eta": ["Distribution of transfers", r"$\eta_{j,s,t}$"],
    "alpha_G": ["Government spending as a share of GDP", r"$\alpha^{G}_{t}$"],
    "tG1": ["Model period in which budget closure rule starts", r"$t_{G1}$"],
    "tG2": ["Model period in which budget closure rule ends", r"$t_{G2}$"],
    "rho_G": ["Budget closure rule smoothing parameter", r"$\rho_{G}$"],
    "debt_ratio_ss": ["Steady-state Debt-to-GDP ratio", r"$\bar{\alpha}_{D}$"],
    "initial_debt_ratio": [
        "Initial period Debt-to-GDP ratio",
        r"$\alpha_{D,0}$",
    ],
    "r_gov_scale": [
        "Scale parameter in government interest rate wedge",
        r"$\tau_{d,t}$",
    ],
    "r_gov_shift": [
        "Shift parameter in government interest rate wedge",
        r"$\mu_{d,t}$",
    ],
    "AIME_num_years": [
        "Number of years over which compute AIME",
        r"$\texttt{AIME_num_years}$",
    ],
    "AIME_bkt_1": ["First AIME bracket threshold", r"$\texttt{AIME_bkt_1}$"],
    "AIME_bkt_2": ["Second AIME bracket threshold", r"$\texttt{AIME_bkt_2}$"],
    "PIA_rate_bkt_1": [
        "First AIME bracket PIA rate",
        r"$\texttt{PIA_rate_bkt_1}$",
    ],
    "PIA_rate_bkt_2": [
        "Second AIME bracket PIA rate",
        r"\texttt{PIA_rate_bkt_2}",
    ],
    "PIA_rate_bkt_3": [
        "Third AIME bracket PIA rate",
        r"$\texttt{PIA_rate_bkt_3}$",
    ],
    "PIA_maxpayment": ["Maximum PIA payment", r"$\texttt{PIA_maxpayment}$"],
    "PIA_minpayment": ["Minimum PIA payment", r"$\texttt{PIA_maxpayment}$"],
    "replacement_rate_adjust": [
        "Adjustment to replacement rate",
        r"$theta_{adj,t}$",
    ],
    "world_int_rate": ["World interest rate", r"$r^{*}_{t}$"],
    "initial_foreign_debt_ratio": [
        "Share of government debt held by foreigners in initial period",
        r"$D_{f,0}$",
    ],
    "zeta_D": [
        "Share of new debt issues purchased by foreigners",
        r"$\zeta_{D, t}$",
    ],
    "zeta_K": [
        "Share of excess capital demand satisfied by foreigners",
        r"$\zeta_{K, t}$",
    ],
    "nu": ["Dampening parameter for TPI", r"$\xi$"],
    "maxiter": ["Maximum number of iterations for TPI", r"$\texttt{maxiter}$"],
    "mindist_SS": ["SS solution tolerance", r"$\texttt{mindist_SS}$"],
    "mindist_TPI": ["TPI solution tolerance", r"$\texttt{mindist_TPI}$"],
}

# Ignoring the following:
# 'starting_age', 'ending_age', 'constant_demographics',
# 'constant_rates', 'zero_taxes'

"""
Create dictionaries to map micro categories to broad groups
"""
CONS_DICT = {
    "Food": [
        "cmaiz",
        "crice",
        "cocer",
        "coils",
        "croot",
        "cvege",
        "csugr",
        "ctoba",
        "ccott",
        "cfrui",
        "ccoff",
        "cocrp",
        "ccatt",
        "cpoul",
        "coliv",
        "cfore",
        "cfish",
        "cfood",
        "cbeve",
    ],
    "Energy and water": ["cmine", "celec", "cwatr"],
    "Non-durables": [
        "ctext",
        "cwood",
        "cchem",
        "cnmet",
        "cmetl",
    ],
    "Durables": [
        "cmach",
        "coman",
        "ccons",
    ],
    "Services": [
        "ctrad",
        "ctran",
        "chotl",
        "ccomm",
        "cfsrv",
        "creal",
        "cbsrv",
        "cpadm",
        "ceduc",
        "cheal",
        "cosrv",
    ],
}

PROD_DICT = {
    "Agriculture and Fishing": [
        "amaiz",
        "arice",
        "aocer",
        "aoils",
        "aroot",
        "avege",
        "asugr",
        "atoba",
        "acoff",
        "afrui",
        "acoff",
        "aocrp",
        "acatt",
        "apoul",
        "aoliv",
        "afore",
        "afish",
    ],
    "Mining": [
        "amine",
    ],
    "Utilities": ["aelec", "awatr"],
    "Construction": ["acons"],
    "Trade and Transport": [
        "atrad",
        "atran",
    ],
    "Services": [
        "ahotl",
        "acomm",
        "afsrv",
        "areal",
        "absrv",
        "apadm",
        "aeduc",
        "aheal",
        "aosrv",
    ],
    "Manufacturing": [
        "afood",
        "abeve",
        "atext",
        "awood",
        "achem",
        "anmet",
        "ametl",
        "amach",
        "aoman",
    ],
}



import numpy as np
import scipy.optimize as opt
import scipy.interpolate as si
from ogcore import parameter_plots as pp
from ogcore import utils
from ogcore.parameters import Specifications
import os
import json
import urllib.request

CUR_PATH = os.path.abspath(os.path.dirname(__file__))
OUTPUT_DIR = os.path.join(CUR_PATH, "OUTPUT", "ability")


def get_e_interp(E, S, J, lambdas, age_wgts, gini_to_match=40.7, plot=False):
    """
    This function takes the calibrated lifetime earnings profiles
    (abilities, e matrix) from OG-USA and then adjusts the shape of those
    profiles to match the Gini coefficient for another economy. The
    Gini coefficient to match is given in the argument gini_to_match.
    Note that the calibrated OG-USA e matrix is of size (80, 10), where
    80 is the number of ages and 10 is the number of ability types.
    Users of this function specify their own number of age groups (S)
    and ability types (J). The function will map the fitted functions
    into these dimensions so long as the percentiles of the ability types
    given in lambdas is not more refined at the top end than those in
    OG-USA (which identifies up to the top 0.1%).

    Args:
        E (int): the age agents become economically active
        S (int): number of ages to interpolate. This method assumes that
            ages are evenly spaced between the beginning of age E
            up to E+S, >= 3
        J (int): number of ability types to interpolate
        lambdas (Numpy array): distribution of population in each
            ability group, length J
        age_wgts (Numpy array): distribution of population in each age
            group, length S
        gini_to_match (float): Gini coefficient to match, default is
            40.7, the Gini coefficient for PHL in 2023
            https://data.worldbank.org/indicator/SI.POV.GINI
        plot (bool): if True, creates plots of emat_orig and the new
            interpolated emat_new

    Returns:
        emat_new_scaled (Numpy array): interpolated ability matrix scaled
            so that population-weighted average is 1, size SxJ

    """
    assert lambdas.shape[0] == J
    assert age_wgts.shape[0] == S
    # Load USA e matrix as a baseline
    usa_params = Specifications()
    usa_params.update_specifications(
        json.load(
            urllib.request.urlopen(
                "https://raw.githubusercontent.com/PSLmodels/OG-USA/master/ogusa/ogusa_default_parameters.json"
            )
        )
    )

    # Define a function that will find the "a" in the equation:
    # e_Y = e_USA * exp(a * e_USA)
    # such that the e_Y produces a gini coefficient in the model that
    # gives the same ratio between the model implied Gini's in the USA
    # and the target country and the empirical Gini's in the USA and given
    # by gin_to_match for the target country
    def f(
        a,
        emat_orig,
        age_wgts,
        abil_wgts,
        gini_to_match,
        gini_usa_data,
        gini_usa_model,
    ):
        gini_target_model = utils.Inequality(
            emat_orig * np.exp(a * emat_orig),
            age_wgts,
            abil_wgts,
            len(age_wgts),
            len(abil_wgts),
        ).gini()
        error = (gini_to_match / gini_usa_data) - (
            gini_target_model / gini_usa_model
        )
        return error

    # Note, USA gini in the World Bank data is 41.5
    # See https://data.worldbank.org/indicator/SI.POV.GINI
    gini_usa_data = 41.5
    # Find the model implied Gini for the USA
    gini_usa_model = utils.Inequality(
        usa_params.e[0, :, :],
        usa_params.omega_SS,
        usa_params.lambdas,
        usa_params.S,
        usa_params.J,
    ).gini()

    x = opt.root_scalar(
        f,
        args=(
            usa_params.e[0, :, :],
            usa_params.omega_SS,
            usa_params.lambdas,
            gini_to_match,
            gini_usa_data,
            gini_usa_model,
        ),
        method="bisect",
        bracket=[-1, 1],
        xtol=1e-10,
    )
    a = x.root
    e_new = usa_params.e[0, :, :] * np.exp(a * usa_params.e[0, :, :])
    emat_new_scaled = (
        e_new
        / (
            e_new
            * usa_params.omega_SS.reshape(usa_params.S, 1)
            * usa_params.lambdas.reshape(1, usa_params.J)
        ).sum()
    )
    # Now interpolate for the cases where S and/or J not the same in the
    # country parameterization as in the default USA parameterization
    if (
        S == usa_params.S
        and np.array_equal(
            usa_params.lambdas,
            lambdas,
        )
        is True
    ):
        pass  # will return the e_new_scaled found above since dims the same
    else:
        # generate vector of mid points for the Filipino ability groups
        abil_midp = np.zeros(J)
        pct_lb = 0.0
        for j in range(J):
            abil_midp[j] = pct_lb + 0.5 * lambdas[j]
            pct_lb += lambdas[j]
        # generate vector of mid points for the USA ability groups
        M = usa_params.lambdas.shape[0]
        emat_j_midp = np.zeros(M)
        pct_lb = 0.0
        for m in range(M):
            emat_j_midp[m] = pct_lb + 0.5 * usa_params.lambdas[m]
            pct_lb += usa_params.lambdas[m]

        # Make sure that values in abil_midp are within interpolating
        # bounds
        if abil_midp.min() < emat_j_midp.min() or abil_midp.max() > (
            1 - usa_params.lambdas[-1]
        ):
            err = (
                "One or more entries in abilities vector (lambdas) is outside the "
                + "allowable bounds for interpolation."
            )
            raise RuntimeError(err)
        usa_step = 80 / usa_params.S
        emat_s_midp = np.linspace(
            usa_params.E + 0.5 * usa_step,
            usa_params.E + usa_params.S - 0.5 * usa_step,
            usa_params.S,
        )
        emat_j_mesh, emat_s_mesh = np.meshgrid(emat_j_midp, emat_s_midp)
        newstep = 80 / S
        new_s_midp = np.linspace(E + 0.5 * newstep, E + S - 0.5 * newstep, S)
        new_j_mesh, new_s_mesh = np.meshgrid(abil_midp, new_s_midp)
        newcoords = np.hstack(
            (
                emat_s_mesh.reshape((usa_params.S * usa_params.J, 1)),
                emat_j_mesh.reshape((usa_params.S * usa_params.J, 1)),
            )
        )
        emat_new = si.griddata(
            newcoords,
            emat_new_scaled.flatten(),
            (new_s_mesh, new_j_mesh),
            method="linear",
        )
        emat_new_scaled = (
            emat_new
            / (emat_new * age_wgts.reshape(S, 1) * lambdas.reshape(1, J)).sum()
        )

        if plot:
            kwargs = {"filesuffix": "_intrp_scaled"}
            pp.plot_income_data(
                new_s_midp,
                abil_midp,
                lambdas,
                emat_new_scaled,
                OUTPUT_DIR,
                **kwargs,
            )

    return emat_new_scaled



import requests
import urllib3
import ssl
import socket


class CustomHttpAdapter(requests.adapters.HTTPAdapter):
    """
    The UN Data Portal server doesn't support
    "RFC 5746 secure renegotiation". This causes and error when the
    client is using OpenSSL 3, which enforces that standard by default.
    The fix is to create a custom SSL context that allows for legacy
    connections. This defines a function get_legacy_session() that
    should be used instead of requests().
    """

    # "Transport adapter" that allows us to use custom ssl_context.
    def __init__(self, ssl_context=None, **kwargs):
        self.ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = urllib3.poolmanager.PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            ssl_context=self.ssl_context,
        )


def get_legacy_session():
    ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    ctx.options |= 0x4  # OP_LEGACY_SERVER_CONNECT  #in Python 3.12 you will be able to switch from 0x4 to ssl.OP_LEGACY_SERVER_CONNECT.
    session = requests.session()
    session.mount("https://", CustomHttpAdapter(ctx))
    return session


# Function to check if connected to internet
def is_connected():
    try:
        # connect to the host -- tells us if the host is actually
        # reachable
        socket.create_connection(("1.1.1.1", 53))
        return True
    except OSError:
        pass
    return False



"""
Specify what is available to import from the ogphl package.
"""

from ogphl.calibrate import *
from ogphl.income import *
from ogphl.input_output import *
from ogphl.macro_params import *
from ogphl.utils import *

__version__ = "0.0.14"



"""
Tests of income.py module
"""

import pytest
import numpy as np
from ogphl import income


# @pytest.mark.parametrize(
#     "abil_wgts,expected_vals",
#     [
#         (abil_wgts1, expected_vals1),
#         (abil_wgts2, expected_vals2),
#         (abil_wgts3, expected_vals3),
#         (abil_wgts4, expected_vals4),
#     ],
#     ids=[
#         "J=7, default weights",
#         "non-default weights",
#         "J=10 weights",
#         "J=9 weights",
#     ],
# )
# Commenting out for now - need to update values above for PHL and
# there's probably a better way to do this
# def test_get_e_interp(abil_wgts, expected_vals):
#     """
#     Test of get_e_interp
#     """
#     age_wgts = np.ones(80) * 1 / 80
#     test_vals = income.get_e_interp(
#         80, age_wgts, age_wgts, abil_wgts, plot=True
#     )
#     print("test vals = ", test_vals)
#     assert np.allclose(test_vals, expected_vals)


def test_get_e_interp_exception():
    """
    Test that RuntimeError is abil_wgts not suitable for interpolation
    """
    age_wgts = np.ones(80) * 1 / 80
    abil_wgts = np.array([0.25, 0.25, 0.2, 0.1, 0.1, 0.09, 0.009999, 0.000001])
    with pytest.raises(RuntimeError):
        income.get_e_interp(20, 80, 8, abil_wgts, age_wgts)



import setuptools

with open("README.md", "r") as readme_file:
    longdesc = readme_file.read()

setuptools.setup(
    name="ogphl",
    version="0.0.14",
    author="Marcelo LaFleur, Richard W. Evans, and Jason DeBacker",
    license="CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
    description="Philippines Calibration for OG-Core",
    long_description_content_type="text/markdown",
    long_description=longdesc,
    keywords="PHL Philippines calibration of large scale overlapping generations model of fiscal policy",
    url="https://github.com/EAPD-DRB/OG-PHL/",
    download_url="https://github.com/EAPD-DRB/OG-PHL/",
    project_urls={
        "Issue Tracker": "https://github.com/EAPD-DRB/OG-PHL/issues",
    },
    packages=["ogphl"],
    package_data={"ogphl": ["ogphl_default_parameters.json", "data/*"]},
    include_packages=True,
    python_requires=">3.11, <3.14",
    install_requires=[
        "numpy",
        "psutil",
        "scipy>=1.7.1",
        "pandas>=1.2.5",
        "matplotlib",
        "dask>=2.30.0",
        "distributed>=2.30.1",
        "paramtools>=0.20.0",
        "requests",
        "pandas-datareader",
        "xlwt",
        "openpyxl>=3.1.2",
        "statsmodels",
        "linearmodels",
        "wheel",
        "black",
        "linecheck",
        "ogcore>=0.14.5",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Common Public License",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    tests_require=["pytest"],
)
